SystemOrganization addCategory: #'Protocols-Kernel'!
SystemOrganization addCategory: #'Protocols-Type Vocabularies'!

Object subclass: #ObjectWithDocumentation
	instanceVariableNames: 'authoringStamp properties elementSymbol naturalLanguageTranslations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!ObjectWithDocumentation commentStamp: '<historical>' prior: 0!
ObjectWithDocumentation - an abstract superclass for objects that allows maintenance of an authoring stamp, a body of documentation, and a properties dictionary.
The Properties implementation has not happened yet -- it would closely mirror the implemenation of properties in the MorphExtension, for example.!

ObjectWithDocumentation subclass: #ElementCategory
	instanceVariableNames: 'categoryName keysInOrder elementDictionary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!ElementCategory commentStamp: '<historical>' prior: 0!
ElementCategory
	
Contains a list of elements that affords keyed access but also has an inherent order.

Add items to the category by sending it elementAt:put:.
Obtain the elements in order by sending #elementsInOrder
Obtain the value of an element at a given key by sending #elementAt:!

!ElementCategory methodsFor: 'initialization' stamp: 'sw 3/30/2001 00:12'!
addCategoryItem: anItem
	"Add the item at the end, obtaining its key from itself (it must respond to #categoryName)"

	self elementAt: anItem categoryName put: anItem! !

!ElementCategory methodsFor: 'category name' stamp: 'sw 1/26/2001 22:45'!
categoryName
	"Answer the formal name of the category"

	^ categoryName! !

!ElementCategory methodsFor: 'category name' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryName: aName
	"Set the category name"

	categoryName := aName! !

!ElementCategory methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
clear
	"Clear the receiber's keysInOrder and elementDictionary"

	keysInOrder := OrderedCollection new.
	elementDictionary := IdentityDictionary new! !

!ElementCategory methodsFor: 'copying' stamp: 'sw 12/1/2000 22:45'!
copy
	"Answer a copy of the receiver"

	^ super copy copyFrom: self! !

!ElementCategory methodsFor: 'copying' stamp: 'stephaneducasse 2/4/2006 20:39'!
copyFrom: donor
	"Copy the receiver's contents from the donor"

	keysInOrder := donor keysInOrder.
	elementDictionary := donor copyOfElementDictionary! !

!ElementCategory methodsFor: 'copying' stamp: 'sw 12/1/2000 22:46'!
copyOfElementDictionary
	"Answer a copy of the element dictionary"

	^ elementDictionary copy! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 12/1/2000 22:46'!
elementAt: aKey
	"Answer the element at the given key"

	^ elementDictionary at: aKey ifAbsent: [nil]! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 1/26/2001 22:54'!
elementAt: sym put: element
	"Add symbol at the end of my sorted list (unless it is already present), and put the element in the dictionary"

	(keysInOrder includes: sym) ifFalse: [keysInOrder add: sym].
	^ elementDictionary at: sym put: element! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 9/12/2001 22:59'!
elementSymbol
	"Answer the element symbol for the receiver.  Here, the categoryName dominates"

	^ categoryName! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 12/1/2000 22:47'!
elementsInOrder
	"Answer the elements in order"

	^ keysInOrder collect: [:aKey | elementDictionary at: aKey]! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 4/3/2001 11:06'!
fasterElementAt: sym put: element
	"Add symbol at the end of my sorted list and put the element in the dictionary.  This variant adds the key at the end of the keys list without checking whether it already exists."

	keysInOrder add: sym.
	^ elementDictionary at: sym put: element! !

!ElementCategory methodsFor: 'keys' stamp: 'sw 12/11/2000 15:36'!
includesKey: aKey
	"Answer whether the receiver's dictionary holds the given key"

	^ elementDictionary includesKey: aKey! !

!ElementCategory methodsFor: 'private' stamp: 'sw 8/6/2004 10:34'!
initWordingAndDocumentation
	"Initialize wording and documentation (helpMessage) for getters and setters"

	self wording: self categoryName! !

!ElementCategory methodsFor: 'initialization' stamp: 'sw 3/28/2001 19:47'!
initialize
	"Initialize the receiver (automatically called when instances are created via 'new')"

	super initialize.
	self clear! !

!ElementCategory methodsFor: 'keys' stamp: 'sw 12/1/2000 22:47'!
keysInOrder
	"Answer the keys in their sorted order"

	^ keysInOrder copy! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 1/26/2001 23:00'!
placeKey: key1 afterKey: key2
	"Place the first key after the second one in my keysInOrder ordering"

	keysInOrder remove: key1.
	keysInOrder add: key1 after: key2! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 1/26/2001 23:00'!
placeKey: key1 beforeKey: key2
	"Place the first key before the second one in my keysInOrder ordering"

	keysInOrder remove: key1.
	keysInOrder add: key1 before: key2! !

!ElementCategory methodsFor: 'printing' stamp: 'sw 1/26/2001 22:47'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that identifies the receiver."

	super printOn: aStream.
	categoryName ifNotNil: [aStream nextPutAll: ' named ', categoryName asString]! !

!ElementCategory methodsFor: 'elements' stamp: 'sw 4/11/2001 20:08'!
removeElementAt: aKey
	"Remove the element at the given key"

	elementDictionary removeKey: aKey ifAbsent: [^ self].
	keysInOrder remove: aKey ifAbsent: []! !

!ElementCategory methodsFor: 'translation' stamp: 'dgd 12/4/2003 20:22'!
translated
	"answer the receiver translated to the current language"
	
	^ self class new categoryName: categoryName asString translated asSymbol! !

ObjectWithDocumentation subclass: #MethodInterface
	instanceVariableNames: 'selector argumentVariables resultSpecification receiverType attributeKeywords defaultStatus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!MethodInterface commentStamp: '<historical>' prior: 0!
A MethodInterface describes the interface for a single method.  The most generic form is not bound to any particular class or object but rather describes an idealized interface.

	selector					A symbol - the selector being described
	argumentSpecifications	A list of specifications for the formal arguments of the method
	resultSpecification 		A characterization of the return value of the method
	userLevel				
	attributeKeywords		A list of symbols, comprising keywords that the user wishes to
								see on the screen for this method
	defaultStatus			The status to apply to new instances of the class by default
							(#ticking, #paused, #normal, etc.)


!

!MethodInterface classSide methodsFor: 'utilities' stamp: 'sw 7/17/2001 19:08'!
firingInterface
	"Answer an instance of the receiver representing #fire"

	^ self new selector: #fire type: nil setter: nil! !

!MethodInterface classSide methodsFor: 'utilities' stamp: 'gk 3/1/2005 10:43'!
isNullMarker: aMarker
	"Answer true if aMarker is nil or is one of the symbols in #(none #nil unused missing) -- to service a variety of historical conventions"

	^ aMarker isNil or: [#(none #nil unused missing) includes: aMarker]

"
MethodInterface isNullMarker: nil
MethodInterface isNullMarker: #nil
MethodInterface isNullMarker: #none
MethodInterface isNullMarker: #znak
"! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
argumentVariables
	"Answer the list of argumentVariables of the interface"

	^ argumentVariables ifNil: [argumentVariables := OrderedCollection new]! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
argumentVariables: variableList
	"Set the argument variables"

	argumentVariables := variableList! !

!MethodInterface methodsFor: 'attribute keywords' stamp: 'stephaneducasse 2/4/2006 20:39'!
attributeKeywords
	"Answer a list of attribute keywords associated with the receiver"

	^ attributeKeywords ifNil: [attributeKeywords := OrderedCollection new]! !

!MethodInterface methodsFor: 'access' stamp: 'sw 3/8/2001 16:29'!
companionSetterSelector
	"If there is a companion setter selector, anwer it, else answer nil"

	^ resultSpecification ifNotNil:
		[resultSpecification companionSetterSelector]! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
conjuredUpFor: aSelector class: aClass
	"Initialize the receiver to have the given selector, obtaining whatever info one can from aClass.  This basically covers the situation where no formal definition has been made."

	| parts |
	self initializeFor: aSelector.
	self wording: aSelector.

	receiverType := #unknown.
	parts := aClass formalHeaderPartsFor: aSelector.
	argumentVariables := (1 to: selector numArgs) collect:
		[:anIndex | Variable new name: (parts at: (4 * anIndex)) type: #Object].
	parts last isEmptyOrNil ifFalse: [self documentation: parts last].
! !

!MethodInterface methodsFor: 'status' stamp: 'sw 1/23/2001 17:16'!
defaultStatus
	"Answer the receiver's default defaultStatus"

	^ defaultStatus! !

!MethodInterface methodsFor: 'status' stamp: 'stephaneducasse 2/4/2006 20:39'!
defaultStatus: aStatus
	"Set the receiver's defaultStatus as indicated"

	defaultStatus := aStatus! !

!MethodInterface methodsFor: 'access' stamp: 'sw 9/13/2001 16:42'!
elementSymbol
	"Answer the element symbol, for the purposes of translation"

	^ selector! !

!MethodInterface methodsFor: 'attribute keywords' stamp: 'sw 1/26/2001 16:32'!
flagAttribute: aSymbol
	"Mark the receiver as having the given category-keyword"

	(self attributeKeywords includes: aSymbol) ifFalse: [attributeKeywords add: aSymbol]! !

!MethodInterface methodsFor: 'attribute keywords' stamp: 'sw 1/26/2001 16:33'!
flagAttributes: attributeSymbolList
	"Mark the receiver has being flagged with all the symbols in the list provided"

	attributeSymbolList do: [:aSym | self flagAttribute: aSym]! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initialize
	"Initialize the receiver"

	super initialize.
	attributeKeywords := OrderedCollection new.
	defaultStatus := #normal.
	argumentVariables := OrderedCollection new
! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initializeFor: aSelector
	"Initialize the receiver to have the given selector"

	selector := aSelector.
	attributeKeywords := OrderedCollection new.
	defaultStatus := #normal
! !

!MethodInterface methodsFor: 'attribute keywords' stamp: 'sw 1/26/2001 16:34'!
isFlaggedAs: aSymbol
	"Answer whether the receiver is flagged with the given trait"

	^ self attributeKeywords includes: aSymbol! !

!MethodInterface methodsFor: 'printing' stamp: 'nk 8/20/2004 09:38'!
printOn: aStream
	"print the receiver on a stream.  Overridden to provide details about wording, selector, result type, and companion setter."

	super printOn: aStream.
	aStream nextPutAll: ' - wording: ';
		print: self wording;
		nextPutAll: ' selector: ';
		print: selector.
	self argumentVariables size > 0 ifTrue:
		[aStream nextPutAll: ' Arguments: '.
		argumentVariables doWithIndex:
			[:aVariable :anIndex | 
				aStream nextPutAll: 'argument #', anIndex printString, ' name = ', aVariable variableName asString, ', type = ', aVariable variableType]].
	resultSpecification ifNotNil:
		[aStream nextPutAll: ' result type = ', resultSpecification resultType asString.
		resultSpecification companionSetterSelector ifNotNil:
			[aStream nextPutAll: ' setter = ', resultSpecification companionSetterSelector asString]]
	! !

!MethodInterface methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:39'!
receiverType
	"Answer the receiver type"

	^ receiverType ifNil: [receiverType := #unknown]! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
receiverType: aType
	"set the receiver type.  Whether the receiverType earns its keep here is not yet well understood.  At the moment, this is unsent"

	receiverType := aType! !

!MethodInterface methodsFor: 'initialize-release' stamp: 'ar 3/3/2001 19:38'!
releaseCachedState
	"Sent by player"! !

!MethodInterface methodsFor: 'access' stamp: 'sw 3/10/2001 00:38'!
resultType
	"Answer the result type"

	^ resultSpecification
		ifNotNil:
			[resultSpecification type]
		ifNil:
			[#unknown]! !

!MethodInterface methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
resultType: aType
	"Set the receiver's resultSpecification to be a ResultType of the given type"

	resultSpecification := ResultSpecification new.
	resultSpecification resultType: aType! !

!MethodInterface methodsFor: 'access' stamp: 'sw 1/23/2001 18:37'!
selector
	"Answer the receiver's selector"

	^ selector! !

!MethodInterface methodsFor: 'attribute keywords' stamp: 'stephaneducasse 2/4/2006 20:39'!
selector: aSelector type: aType setter: aSetter
	"Set the receiver's fields as indicated.  Values of nil or #none for the result type and the setter indicate that there is none"

	selector := aSelector.
	(MethodInterface isNullMarker: aType) ifFalse:
		[resultSpecification := ResultSpecification new.
		resultSpecification resultType: aType.
		(MethodInterface isNullMarker: aSetter) ifFalse:
			[resultSpecification companionSetterSelector: aSetter]]! !

!MethodInterface methodsFor: 'initialization' stamp: 'sw 5/26/2001 22:59'!
setNotToRefresh
	"Set the receiver up not to do periodic refresh."

	resultSpecification ifNotNil: [resultSpecification refetchFrequency: nil]! !

!MethodInterface methodsFor: 'initialization' stamp: 'sw 5/3/2001 15:59'!
setToRefetch
	"Set the receiver up to expect a refetch, assuming it has a result specification"

	resultSpecification ifNotNil: [resultSpecification refetchFrequency: 1]! !

!MethodInterface methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:39'!
typeForArgumentNumber: anArgumentNumber
	"Answer the data type for the given argument number"

	| aVariable |
	aVariable := self argumentVariables at: anArgumentNumber.
	^ aVariable variableType! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 21:21'!
documentation
	"Answer the receiver's documentation"

	^self helpMessage! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 19:33'!
documentation: somethingUsefulHopefully
	"Set the receiver's documentation, in the current langauge"

	self helpMessage: somethingUsefulHopefully! !

!ObjectWithDocumentation methodsFor: 'documentation' stamp: 'rbb 3/1/2005 11:02'!
editDescription
	"Allow the user to see and edit the documentation for this object"
	| reply helpMessage |
	helpMessage := self documentation isNil
				ifTrue: [String new]
				ifFalse: [self documentation].
	reply := UIManager default
				multiLineRequest: 'Kindly edit the description' translated
				centerAt: Sensor cursorPoint
				initialAnswer: helpMessage
				answerHeight: 200.
	reply isEmptyOrNil
		ifFalse: [self documentation: reply]! !

!ObjectWithDocumentation methodsFor: 'miscellaneous' stamp: 'sw 9/12/2001 23:03'!
elementSymbol
	"Answer the receiver's element symbol"

	^ elementSymbol! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 21:28'!
getterSetterHelpMessage
	"Returns a helpMessage that has been computed previously and needs to be translated and then formatted with the elementSymbol.
	'get value of {1}' translated format: {elSym}"

	^(self propertyAt: #getterSetterHelpMessage ifAbsent: [^nil])
		translated format: {self elementSymbol}! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 21:29'!
getterSetterHelpMessage: aString
	"Sets a helpMessage that needs to be translated and then formatted with the elementSymbol.
	'get value of {1}' translated format: {elSym}"

	self propertyAt: #getterSetterHelpMessage put: aString! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'sw 8/18/2004 20:23'!
helpMessage
	"Check if there is a getterSetterHelpMessage. 
	Otherwise try the normal help message or return nil."

	^ self getterSetterHelpMessage
		ifNil: [(self propertyAt: #helpMessage ifAbsent:
			[self legacyHelpMessage ifNil: [^ nil]]) translated]! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 19:32'!
helpMessage: somethingUsefulHopefully
	"Set the receiver's documentation, in the current langauge"

	self propertyAt: #helpMessage put: somethingUsefulHopefully! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 21:31'!
initWordingAndDocumentation
	"Initialize wording and documentation (helpMessage) for getters and setters"

	| elSym |
	elSym := self elementSymbol.
	elSym
		ifNil: [^self].

	((elSym beginsWith: 'get')
		and: [elSym size > 3])
		ifTrue: [
			self wording: (elSym allButFirst: 3) withFirstCharacterDownshifted.
			self getterSetterHelpMessage: 'get value of {1}']
		ifFalse: [
			((elSym beginsWith: 'set')
				and: [elSym size > 4])
				ifTrue: [
					self wording: (elSym allButFirst: 3) withFirstCharacterDownshifted.
					self getterSetterHelpMessage: 'set value of {1}']]! !

!ObjectWithDocumentation methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initialize
	"Initialize the receiver (automatically called when instances are created via 'new')"

	authoringStamp := Utilities changeStampPerSe
! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:39'!
legacyHelpMessage
	"If I have a help message stashed in my legacy naturalTranslations slot, answer its translated rendition, else answer nil.  If I *do* come across a legacy help message, transfer it to my properties dictionary."

	| untranslated |
	naturalLanguageTranslations isEmptyOrNil  "only in legacy (pre-3.8) projects"
		ifTrue: [^ nil].
	untranslated := naturalLanguageTranslations first helpMessage ifNil: [^ nil].
	self propertyAt: #helpMessage put: untranslated.
	naturalLanguageTranslations removeFirst.
	naturalLanguageTranslations isEmpty ifTrue: [naturalLanguageTranslations := nil].
	^ untranslated translated! !

!ObjectWithDocumentation methodsFor: 'migration' stamp: 'mir 7/12/2004 23:45'!
migrateWordAndHelpMessage
	"Migrate the English wording and help message to the new structure"

	| englishElement |
	self initWordingAndDocumentation.
	(self properties includes: #wording)
		ifFalse: [
			englishElement := self naturalLanguageTranslations
				detect: [:each | each language == #English] ifNone: [^nil].
			self wording: englishElement wording.
			self helpMessage: englishElement helpMessage]! !

!ObjectWithDocumentation methodsFor: 'miscellaneous' stamp: 'mir 7/12/2004 23:20'!
naturalLanguageTranslations
	^naturalLanguageTranslations ifNil: [OrderedCollection new]! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 19:30'!
properties
	^properties ifNil: [properties := Dictionary new]! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 19:30'!
propertyAt: key
	^self propertyAt: key ifAbsent: [nil]! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 19:29'!
propertyAt: key ifAbsent: aBlock
	^properties
		ifNil: aBlock
		ifNotNil: [properties at: key ifAbsent: aBlock]! !

!ObjectWithDocumentation methodsFor: 'private' stamp: 'mir 7/12/2004 19:29'!
propertyAt: key put: aValue
	self properties at: key put: aValue! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 23:57'!
untranslatedHelpMessage
	"Check if there is a getterSetterHelpMessage. 
	Otherwise try the normal help message or return nil."

	^(self propertyAt: #getterSetterHelpMessage ifAbsent: [nil])
		ifNil: [(self propertyAt: #helpMessage ifAbsent: [nil])]! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 23:56'!
untranslatedWording
	"Answer the receiver's wording"

	^self propertyAt: #wording ifAbsent: [nil]! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 21:34'!
wording
	"Answer the receiver's wording"

	| wording |
	(wording := self propertyAt: #wording ifAbsent: [nil])
		ifNotNil: [^wording translated].

	self initWordingAndDocumentation.
	^self propertyAt: #wording ifAbsent: ['']! !

!ObjectWithDocumentation methodsFor: 'accessing' stamp: 'mir 7/12/2004 21:39'!
wording: aString
	"Set the receiver's wording, in the current langauge"

	self propertyAt: #wording put: aString! !

ObjectWithDocumentation subclass: #ResultSpecification
	instanceVariableNames: 'type companionSetterSelector refetchFrequency'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!ResultSpecification methodsFor: 'companion setter' stamp: 'sw 2/27/2001 09:15'!
companionSetterSelector
	"Answer the companion setter, nil if none"

	^ companionSetterSelector! !

!ResultSpecification methodsFor: 'companion setter' stamp: 'stephaneducasse 2/4/2006 20:39'!
companionSetterSelector: aSetterSelector
	"Set the receiver's companionSetterSelector as indicated"

	companionSetterSelector := aSetterSelector! !

!ResultSpecification methodsFor: 'refetch' stamp: 'sw 5/3/2001 00:07'!
refetchFrequency
	"Answer the frequency with which the receiver should be refetched by a readout polling values from it, as in a Viewer.  Answer nil if not ever to be refetched automatically"

	^ refetchFrequency! !

!ResultSpecification methodsFor: 'refetch' stamp: 'stephaneducasse 2/4/2006 20:39'!
refetchFrequency: aFrequency
	"Set the refetch frequency"

	refetchFrequency := aFrequency! !

!ResultSpecification methodsFor: 'result type' stamp: 'sw 2/27/2001 09:14'!
resultType
	"Answer the reciever's result type"

	^ type! !

!ResultSpecification methodsFor: 'result type' stamp: 'stephaneducasse 2/4/2006 20:39'!
resultType: aType
	"Set the receiver's resultType as specified"

	type := aType! !

!ResultSpecification methodsFor: 'result type' stamp: 'sw 2/24/2001 12:12'!
type
	"Answer the reciever's type"

	^ type! !

ObjectWithDocumentation subclass: #Variable
	instanceVariableNames: 'defaultValue floatPrecision variableName variableType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!Variable methodsFor: 'value' stamp: 'sw 12/11/2000 14:06'!
defaultValue
	"Answer the default value to be supplied to the receiver"

	^ defaultValue! !

!Variable methodsFor: 'name' stamp: 'stephaneducasse 2/4/2006 20:39'!
name: aName type: aType
	"Set the variable's name and type as indicated"

	variableName := aName.
	variableType := aType! !

!Variable methodsFor: 'name' stamp: 'sw 9/30/2004 05:02'!
printOn: aStream
	"Print the receiver on the stream"

	super printOn: aStream.
	aStream nextPutAll: ' named ', (self variableName ifNil: ['<unnamed>']), ' type = ', variableType printString, ' default val = ', defaultValue printString! !

!Variable methodsFor: 'value' stamp: 'stephaneducasse 2/4/2006 20:39'!
sample
	"The closest we can come to an object for our type"

	| ty clsName |
	self defaultValue ifNotNil: [^ self defaultValue].
	ty := self variableType.
	"How translate a type like #player into a class?"
	clsName := ty asString.
	clsName at: 1 put: (clsName first asUppercase).
	clsName := clsName asSymbol.
	(Smalltalk includesKey: clsName) ifFalse: [self error: 'What type is this?'. ^ 5].
	^ (Smalltalk at: clsName) initializedInstance! !

!Variable methodsFor: 'name' stamp: 'sw 12/11/2000 14:07'!
variableName
	"Answer the variable name of the receiver"

	^ variableName! !

!Variable methodsFor: 'type' stamp: 'sw 12/11/2000 14:06'!
variableType
	"Anser the variable type of the receiver"

	^ variableType! !

!Variable methodsFor: 'type' stamp: 'stephaneducasse 2/4/2006 20:39'!
variableType: aType
	"Set the receiver's variable type as requested"

	variableType := aType! !

ObjectWithDocumentation subclass: #Vocabulary
	instanceVariableNames: 'vocabularyName categories methodInterfaces object limitClass translationTable'
	classVariableNames: 'AllStandardVocabularies LanguageSymbols LanguageTable'
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!Vocabulary commentStamp: '<historical>' prior: 0!
Vocabulary

vocabularyName	a Symbol -- the formal name by which this vocabulary is known.
categories			a list of ElementCategory objects: the categories that comprise the vocabulary
methodInterfaces 	an IdentityDictionary; keys are method selectors, values are MethodInterfaces

A Vocabulary can be either be *abstract*, i.e. associated with a *class*, or it can be *bound to an instance*.  The form that is bound to an instance is still in its infancy however.

object				in the *bound* form, an actual object is associated with the vocabulary
limitClass			in the *bound* form, an actual limit class is assocaited with the vocabulary

AllMethodInterfaces	This class variable is available to hold on to all method interfaces
						defined in the system, regardless of class.  Not deployed in 
						the first version of this code to be circulated externally.

AllVocabularies		A dictionary associating symbols with actual abstract vocabulary instances

------
Hints on making a vocabulary for a new foreign language.  You need build a method like #addGermanVocabulary.  Execute

	Transcript show: Vocabulary eToyVocabulary strings.

and copy the text from the transcript to the method you are building.
	A cheap trick to get started is to use a web site that translates.  Here is how.  For an entry like:  

(clearTurtleTrails			'clear pen trails'			'Clear all the pen trails in the interior.')

	substitute exclamation points for single quotes by using Alt-shift-j.  Most of the statements are imperatives.

(clearOwnersPenTrails !!clear all pen trails!! !!clear all pen trails in my containing play field!!) 

	This translates to

(ClearOwnersPenTrails!! reinigt allen Kugelschreiber verfolgt!! !! Reinigt allen Kugelschreiber Pfade in meinem enthaltenden Spiel Feld sind!!) 

	Which is laughable, and will certainly stimulate the author to improve it.
------!

Vocabulary subclass: #DataType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Type Vocabularies'!

!DataType commentStamp: 'sw 8/22/2002 15:01' prior: 0!
A Vocabulary representing typed data.!

!DataType methodsFor: 'initial value' stamp: 'sw 9/26/2001 12:00'!
initialValueForASlotFor: aPlayer
	"Answer the value to give initially to a newly created slot of the given type in the given player"

	^ 'no value'! !

!DataType methodsFor: 'queries' stamp: 'mir 7/15/2004 10:34'!
representsAType
	"Answer whether this vocabulary represents an end-user-sensible data type"

	"^ (self class == DataType) not"  "i.e. subclasses yes, myself no"
	"Assuming this is an abstract class"
	^true! !

DataType subclass: #FullVocabulary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!FullVocabulary commentStamp: '<historical>' prior: 0!
The vocabulary that it all-encompassing.  Its categories consist of the union of all categories of a class and all its superclasses.  The methods in each category consist of those with selectors that are associated with that category.!

!FullVocabulary methodsFor: 'method list' stamp: 'stephaneducasse 2/4/2006 20:39'!
allMethodsInCategory: categoryName forInstance: anObject ofClass: aClass
	"Answer a list of all methods which are in the given category, on behalf of anObject"

	| classToUse |
	classToUse := aClass ifNil: [anObject class].
	^ classToUse allMethodsInCategory: categoryName! !

!FullVocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoriesContaining: aSelector forClass: aTargetClass 
	"Answer a list of category names (all symbols) of categories that contain 
	the given selector for the target object. Initially, this just returns one."
	| classDefiningSelector catName |
	classDefiningSelector := aTargetClass whichClassIncludesSelector: aSelector.
	classDefiningSelector
		ifNil: [^ OrderedCollection new].
	catName := classDefiningSelector whichCategoryIncludesSelector: aSelector.
	^ OrderedCollection with: catName! !

!FullVocabulary methodsFor: 'category list' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryListForInstance: anObject ofClass: aClass limitClass: mostGenericClass
	"Answer the category list for the given object, considering only code implemented in mostGeneric and lower (or higher, depending on which way you're facing"

	| classToUse |
	classToUse := anObject ifNil: [aClass] ifNotNil: [anObject class].
	^ mostGenericClass == classToUse
		ifTrue:
			[mostGenericClass organization categories]
		ifFalse:
			[classToUse allMethodCategoriesIntegratedThrough: mostGenericClass]! !

!FullVocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryWithNameIn: categoryNames thatIncludesSelector: aSelector forInstance: targetInstance ofClass: targetClass 
	"Answer the name of a category, from among the provided 
	categoryNames, which defines the selector for the given class. Here, if 
	the category designated by the implementing class is acceptable it is the 
	one returned"
	| aClass catName result |
	(aClass := targetClass whichClassIncludesSelector: aSelector)
		ifNotNil: [(categoryNames includes: (catName := aClass whichCategoryIncludesSelector: aSelector))
				ifTrue: [catName ~~ #'as yet unclassified'
						ifTrue: [^ catName]]].
	result := super
				categoryWithNameIn: categoryNames
				thatIncludesSelector: aSelector
				forInstance: targetInstance
				ofClass: aClass.
	^ result
		ifNil: [#'as yet unclassified']! !

!FullVocabulary methodsFor: 'queries' stamp: 'sw 12/12/2000 06:06'!
encompassesAPriori: aClass
	"Answer whether an object, by its very nature, is one that the receiver embraces"

	^ true! !

!FullVocabulary methodsFor: 'queries' stamp: 'sw 3/20/2001 15:42'!
includesDefinitionForSelector: aSelector
	"Answer whether the given selector is known to the vocabulary.  Unsent at the moment, may disappear."

	^ true! !

!FullVocabulary methodsFor: 'queries' stamp: 'sw 12/1/2000 21:57'!
includesSelector: aSelector
	"Answer whether the given selector is known to the vocabulary"

	^ true! !

!FullVocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
includesSelector: aSelector forInstance: anInstance ofClass: aTargetClass limitClass: mostGenericClass
	"Answer whether the vocabulary includes the given selector for the given class, only considering method implementations in mostGenericClass and lower"

	| classToUse aClass |
	classToUse := self classToUseFromInstance: anInstance ofClass: aTargetClass.
	^ (aClass := classToUse whichClassIncludesSelector: aSelector)
		ifNil:
			[false]
		ifNotNil:
			[aClass includesBehavior: mostGenericClass]! !

!FullVocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initialize
	"Initialize the receiver (automatically called when instances are created via 'new')
Vocabulary initialize
"

	super initialize.
	vocabularyName := #Object.
	self documentation: '"Object" is all-encompassing vocabulary that embraces all methods understood by an object'.
	self rigAFewCategories! !

!FullVocabulary methodsFor: 'queries' stamp: 'sw 9/27/2001 03:28'!
representsAType
	"Answer whether this vocabulary represents an end-user-sensible data type"

	^ false! !

!FullVocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
rigAFewCategories
	"Formerly used to rig generic categories, now seemingly disfunctional and in abeyance"

	| aMethodCategory |
	true ifTrue: [^ self].

	self flag: #deferred.
"Vocabulary fullVocabulary rigAFewCategories "
	#(	(accessing	'Generally holds methods to read and write instance variables')
		(initialization	'messages typically sent when an object is created, to set up its initial state'))

		do:
			[:pair |
				aMethodCategory := ElementCategory new categoryName: pair first.
				aMethodCategory documentation: pair second.
				self addCategory: aMethodCategory]! !

FullVocabulary subclass: #ScreenedVocabulary
	instanceVariableNames: 'methodScreeningBlock categoryScreeningBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Protocols-Kernel'!

!ScreenedVocabulary methodsFor: 'enumeration' stamp: 'sw 1/5/2001 06:55'!
allMethodsInCategory: categoryName forInstance: anObject ofClass: aClass
	"Answer a list of all methods in the vocabulary which are in the given category, on behalf of the given class and object"

	^ (super allMethodsInCategory: categoryName forInstance: anObject ofClass: aClass) select:
		[:aSelector | self includesSelector: aSelector]! !

!ScreenedVocabulary methodsFor: 'enumeration' stamp: 'sw 12/14/2000 14:03'!
categoryListForInstance: anObject ofClass: aClass limitClass: mostGenericClass
	"Answer the category list for the given object/class, considering only code implemented in mostGenericClass and lower"

	^ (super categoryListForInstance: anObject ofClass: aClass limitClass: mostGenericClass) select:
		[:aCategory | categoryScreeningBlock value: aCategory]! !

!ScreenedVocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryScreeningBlock: aBlock
	"Set the receiver's categoryScreeningBlock to the block provided"

	categoryScreeningBlock := aBlock! !

!ScreenedVocabulary methodsFor: 'queries' stamp: 'sw 12/14/2000 14:06'!
includesSelector: aSelector
	"Answer whether the given selector is known to the vocabulary"

	^ methodScreeningBlock value: aSelector! !

!ScreenedVocabulary methodsFor: 'queries' stamp: 'sw 12/14/2000 06:01'!
includesSelector: aSelector forInstance: anInstance ofClass: aTargetClass limitClass: mostGenericClass
	"Answer whether the vocabulary includes the given selector for the given object, only considering method implementations in mostGenericClass and lower"

	^ (super includesSelector: aSelector forInstance: anInstance ofClass: aTargetClass limitClass: mostGenericClass) and:
		[self includesSelector: aSelector]! !

!ScreenedVocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initialize
	"Initialize the receiver (automatically called when instances are created via 'new')"

	super initialize.
	vocabularyName :=  #Public.
	self documentation: '"Public" is vocabulary that excludes categories that start with "private" and methods that start with "private" or "pvt"'! !

!ScreenedVocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
methodScreeningBlock: aBlock
	"Set the receiver's methodScreeningBlock to the block provided"

	methodScreeningBlock := aBlock! !

!Vocabulary classSide methodsFor: 'standard vocabulary access' stamp: 'mir 7/15/2004 10:55'!
addStandardVocabulary: aVocabulary
	"Add a vocabulary to the list of standard vocabularies"

	self allStandardVocabularies at: aVocabulary vocabularyName put: aVocabulary! !

!Vocabulary classSide methodsFor: 'standard vocabulary access' stamp: 'stephaneducasse 2/4/2006 20:39'!
allStandardVocabularies
	"Answer a list of the currently-defined vocabularies in my AllStandardVocabularies list"
	"Vocabulary allStandardVocabularies"

	^AllStandardVocabularies ifNil: [AllStandardVocabularies := IdentityDictionary new].

! !

!Vocabulary classSide methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
embraceAddedTypeVocabularies
	"If there are any type-vocabulary subclases not otherwise accounted for, acknowledge them at this time"

	| vocabulary |
	DataType allSubclasses do:
		[:dataType |
			vocabulary := dataType new.
			vocabulary representsAType
				ifTrue: [(self allStandardVocabularies includesKey: vocabulary vocabularyName)
					ifFalse: 	[self addStandardVocabulary: vocabulary]]]! !

!Vocabulary classSide methodsFor: 'universal vocabularies' stamp: 'mir 7/15/2004 10:52'!
fullVocabulary
	"Answer the full vocabulary in my AllStandardVocabularies list, creating it if necessary"

	^ self allStandardVocabularies at: #Full ifAbsentPut: [FullVocabulary new]! !

!Vocabulary classSide methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
instanceWhoRespondsTo: aSelector 
	"Find the most likely class that responds to aSelector. Return an instance 
	of it. Look in vocabularies to match the selector."
	"Most eToy selectors are for Players"
	| mthRefs |
	((self vocabularyNamed: #eToy)
			includesSelector: aSelector)
		ifTrue: [aSelector == #+
				ifFalse: [^ Player new costume: Morph new]].
	"Numbers are a problem"
	((self vocabularyNamed: #Number)
			includesSelector: aSelector)
		ifTrue: [^ 1].
	"Is a Float any different?"
	"String Point Time Date"
	#()
		do: [:nn | ((self vocabularyNamed: nn)
					includesSelector: aSelector)
				ifTrue: ["Ask Scott how to get a prototypical instance"
					^ (Smalltalk at: nn) new]].
	mthRefs := self systemNavigation allImplementorsOf: aSelector.
	"every one who implements the selector"
	mthRefs
		sortBlock: [:a :b | (Smalltalk at: a classSymbol) allSuperclasses size < (Smalltalk at: b classSymbol) allSuperclasses size].
	mthRefs size > 0
		ifTrue: [^ (Smalltalk at: mthRefs first classSymbol) new].
	^ Error new! !

!Vocabulary classSide methodsFor: 'type vocabularies' stamp: 'Lukas Renggli 10/29/2008 10:55'!
newNumberVocabulary
	"Answer a Vocabulary object representing the Number vocabulary to the list of AllVocabularies"

	^ nil! !

!Vocabulary classSide methodsFor: 'universal vocabularies' stamp: 'stephaneducasse 2/4/2006 20:39'!
newPublicVocabulary
	| aVocabulary |
	"Answer a public vocabulary"

	aVocabulary := ScreenedVocabulary new.
	aVocabulary vocabularyName: #Public.
	aVocabulary documentation: '"Public" is vocabulary that excludes categories that start with "private" and methods that start with "private" or "pvt"'.

	aVocabulary categoryScreeningBlock: [:aCategoryName | (aCategoryName beginsWith: 'private') not].
	aVocabulary methodScreeningBlock: [:aSelector | 
		((aSelector beginsWith: 'private') or: [aSelector beginsWith: 'pvt']) not].
	^ aVocabulary
! !

!Vocabulary classSide methodsFor: 'testing and demo' stamp: 'stephaneducasse 2/4/2006 20:39'!
newQuadVocabulary
	"Answer a Quad vocabulary -- something to mess with, to illustrate and explore ideas.  Applies to Quadrangles only."

	| aVocabulary  |
	aVocabulary := Vocabulary new vocabularyName: #Quad.
	aVocabulary documentation: 'A highly restricted test vocabulary that can be used with Quadrangle objects'.
	aVocabulary initializeFromTable:  #(
(borderColor borderColor: () Color (basic color) 'The color of the border' unused updating)
(borderWidth borderWidth: () Number (basic geometry) 'The width of the border' unused updating)
(insideColor insideColor: () Color (basic color) 'The color of the quadrangle' unused updating)
(display none () none (basic display) 'Display the quadrangle directly on the screen')
(width none () Number (geometry) 'The width of the object' unused updating)
(left setLeft: () Number (geometry) 'The left edge' unused updating)
(right setRight: () Number (geometry) 'The right edge' unused updating)
(width setWidth: () Number (geometry) 'The width of the object' unused updating)
(height setHeight: () Number (geometry) 'The height of the object' unused updating)
(hasPositiveExtent none () Boolean (tests) 'Whether the corner is to the lower-right of the origin' unused updating)
(isTall none () Boolean (tests) 'Whether the height is greater than the width' unused updating)).

	^ aVocabulary

"Vocabulary initialize"
"Quadrangle exampleInViewer"! !

!Vocabulary classSide methodsFor: 'etoy vocabularies' stamp: 'stephaneducasse 2/4/2006 20:39'!
newSystemVocabulary
	"Answer a Vocabulary object representing significant requests one can make to the Smalltalk object"

	| aVocabulary |
	aVocabulary := self new.

	aVocabulary vocabularyName: #System.
	aVocabulary documentation: 'Useful messages you might want to send to the current Smalltalk image'.
	aVocabulary initializeFromTable:  #(
(aboutThisSystem none () none (basic queries) 'put up a message describing the system' unused)
(saveAsNewVersion none () none (services) 'advance to the next available image-version number and save the image under that new name' unused znak)
(datedVersion none () String (queries) 'the version of the Squeak system')
(endianness none () String (queries) 'big or little - the byte-ordering of the hardware Squeak is currently running on')
(exitToDebugger none () none (dangerous) 'exits to the host debugger.  Do not use this -- I guarantee you will be sorry.')
(bytesLeft none () Number (basic services) 'perform a garbage collection and answer the number of bytes of free space remaining in the system')
"(browseAllCallsOn: none ((aSelector String)) none (#'queries') 'browse all calls on a selector')
(browseAllImplementorsOf: none ((aSelector String)) none (#'queries') 'browse all implementors of a selector')"

"(allMethodsWithSourceString:matchCase: none ((aString String) (caseSensitive Boolean)) none (queries) 'browse all methods that have the given source string, making the search case-sensitive or not depending on the argument provided.')

(browseMethodsWithString:matchCase: none ((aString String) (caseSensitive Boolean)) none (queries) 'browse all methods that contain the given string in any string literal, making the search case-sensitive or not depending on the argument provided.')

(browseAllImplementorsOf:localTo: none ((aSelector String) (aClass Class)) none (#'queries') 'browse all implementors of a selector that are local to a class')"

).
"(isKindOf: none 	((aClass Class)) Boolean (#'class membership') 'answer whether the receiver''s superclass chain includes aClass')"
	^ aVocabulary

"Vocabulary initialize"
"Vocabulary addStandardVocabulary: Vocabulary newSystemVocabulary"

"SmalltalkImage current basicInspect"
"SmalltalkImage current beViewed"
! !

!Vocabulary classSide methodsFor: 'testing and demo' stamp: 'stephaneducasse 2/4/2006 20:39'!
newTestVocabulary
	"Answer a Test vocabulary -- something to mess with, to illustrate and explore ideas."

	| aVocabulary  |
	aVocabulary := Vocabulary new vocabularyName: #Test.
	aVocabulary documentation: 'An illustrative vocabulary for testing'.
	aVocabulary initializeFromTable:  #(
(isKindOf: none 	((aClass Class)) Boolean (#'class membership') 'answer whether the receiver''s superclass chain includes aClass')
(class none none Class (#'class membership' wimpy) 'answer the the class to which the receiver belongs')
(respondsTo: none ((aSelector Symbol))	Boolean (#'class membership') 'answer whether the receiver responds to the given selector')
(as:	none ((aClass Class)) Object (conversion) 'answer the receiver converted to be a member of aClass')).

	^ aVocabulary
"
	#((#'class membership' 	'Whether an object can respond to a given message, etc.' 	(isKindOf: class respondsTo:))
	(conversion 			'Messages to convert from one kind of object to another' 		(as:  asString))
	(copying				'Messages for making copies of objects'						(copy copyFrom:))
	(equality 				'Testing whether two objects are equal' 						( = ~= == ~~))
	(dependents				'Support for dependency notification'						(addDependent: removeDependent: release))) do:

		[:item | 
			aMethodCategory := ElementCategory new categoryName: item first.
			aMethodCategory documentation: item second.
			item third do:
				[:aSelector | 
					aMethodInterface := MethodInterface new initializeFor: aSelector.
					aVocabulary atKey: aSelector putMethodInterface: aMethodInterface.
					aMethodCategory elementAt: aSelector put: aMethodInterface].
			aVocabulary addCategory: aMethodCategory]."
! !

!Vocabulary classSide methodsFor: 'standard vocabulary access' stamp: 'stephaneducasse 2/4/2006 20:39'!
newTimeVocabulary
	"Answer a Vocabulary object representing me" 
	| aVocabulary aMethodCategory aMethodInterface |
	"Vocabulary newTimeVocabulary"
	"Vocabulary addStandardVocabulary: Vocabulary newTimeVocabulary"

	aVocabulary := self new vocabularyName: #Time.
	aVocabulary documentation: 'Time knows about hours, minutes, and seconds.  For long time periods, use Date'.

#((accessing 			'The basic info'
		(hours minutes seconds))
(arithmetic 				'Basic numeric operations'
		(addTime: subtractTime: max: min: min:max:))
(comparing				'Determining which is larger'
		(= < > <= >= ~= between:and:))
(testing 				'Testing'
		(ifNil: ifNotNil:))
(printing 				'Return a string for this Time'
		(hhmm24 print24 intervalString printMinutes printOn:))
(converting 			'Converting it to another form'
		(asSeconds asString))
(copying 				'Make another one like me'
		(copy))
) do: [:item | 
			aMethodCategory := ElementCategory new categoryName: item first.
			aMethodCategory documentation: item second.
			item third do:
				[:aSelector | 
					aMethodInterface := MethodInterface new initializeFor: aSelector.
					aVocabulary atKey: aSelector putMethodInterface: aMethodInterface.
					aMethodCategory elementAt: aSelector put: aMethodInterface].
			aVocabulary addCategory: aMethodCategory].
	#(#addTime: subtractTime: max: min: = < > <= >= ~= ) do: [:sel |
		(aVocabulary methodInterfaceAt: sel ifAbsent: [self error: 'fix this method']) 
			argumentVariables: (OrderedCollection with:
				(Variable new name: nil type: aVocabulary vocabularyName))].
	^ aVocabulary! !

!Vocabulary classSide methodsFor: 'type vocabularies' stamp: 'mir 7/15/2004 10:31'!
numberVocabulary
	"Answer the standard vocabulary representing numbers, creating it if necessary"

	^self allStandardVocabularies at: #Number ifAbsentPut: [self newNumberVocabulary]! !

!Vocabulary classSide methodsFor: 'testing and demo' stamp: 'mir 7/15/2004 10:53'!
quadVocabulary
	"Answer the Quad vocabulary lurking in my AllStandardVocabularies list, creating it if necessary"
	"Vocabulary quadVocabulary"

	^ self allStandardVocabularies at: #Quad ifAbsentPut: [self newQuadVocabulary]! !

!Vocabulary classSide methodsFor: 'testing and demo' stamp: 'mir 7/15/2004 10:54'!
testVocabulary
	"Answer the Test vocabulary lurking in my AllStandardVocabularies list, creating it if necessary"
	"Vocabulary testVocabulary"

	^ self allStandardVocabularies at: #Test ifAbsentPut: [self newTestVocabulary]! !

!Vocabulary classSide methodsFor: 'type vocabularies' stamp: 'stephane.ducasse 10/28/2008 18:16'!
typeChoices
	"Answer a list of all user-choosable data types"

	| aList |
	(aList := self allStandardVocabularies
		select:
			[:aVocab | aVocab representsAType]
		thenCollect:
			[:aVocab | aVocab vocabularyName]).
	^ aList asSortedArray! !

!Vocabulary classSide methodsFor: 'type vocabularies' stamp: 'stephaneducasse 2/4/2006 20:39'!
vocabularyForClass: aClass
	"Answer the standard vocabulary for that class.  Create it if not present and init message exists.  Answer nil if none exists and no init message present."

	| initMsgName newTypeVocab |
	(self allStandardVocabularies includesKey: aClass name)
		ifTrue: [^self allStandardVocabularies at: aClass name].

	initMsgName := ('new', aClass name, 'Vocabulary') asSymbol.
	^(self respondsTo: initMsgName)
		 ifTrue:	[
			newTypeVocab := self perform: initMsgName.
			self addStandardVocabulary: newTypeVocab.
			newTypeVocab]
		ifFalse: [nil]! !

!Vocabulary classSide methodsFor: 'type vocabularies' stamp: 'stephaneducasse 2/4/2006 20:39'!
vocabularyForType: aType
	"Answer a vocabulary appropriate to the given type, which is normally going to be a symbol such as #Number or #Color.  Answer the Unknown vocabulary as a fall-back"

	| ucSym |
	(aType isKindOf: Vocabulary) ifTrue: [^ aType].
	ucSym := aType capitalized asSymbol.
	^self allStandardVocabularies at: ucSym ifAbsent: [self vocabularyNamed: #unknown]! !

!Vocabulary classSide methodsFor: 'standard vocabulary access' stamp: 'sw 6/4/2001 16:33'!
vocabularyFrom: aNameOrVocabulary
	"Answer the standard vocabulary of the given name, or nil if none found,  For backward compatibilitythe parameter might be an actual vocabulary, in which case return it"

	(aNameOrVocabulary isKindOf: Vocabulary) ifTrue: [^ aNameOrVocabulary].
	^ self vocabularyNamed: aNameOrVocabulary! !

!Vocabulary classSide methodsFor: 'standard vocabulary access' stamp: 'nk 8/22/2004 09:24'!
vocabularyNamed: aName
	"Answer the standard vocabulary of the given name, or nil if none found"

	^ self allStandardVocabularies at: aName asSymbol ifAbsent: []! !

!Vocabulary methodsFor: 'initialization' stamp: 'sw 12/11/2000 15:31'!
addCategory: aCategory
	"Add the given category to my categories list"

	categories add: aCategory! !

!Vocabulary methodsFor: 'initialization' stamp: 'sw 1/26/2001 23:03'!
addCategoryNamed: aCategoryName
	"Add a category of the given name to my categories list,"

	categories add: (ElementCategory new categoryName: aCategoryName asSymbol)! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
addFromTable: aTable
	"Add each method-specification tuples, each of the form:
		(1)	selector
		(2)	companion setter selector (#none or nil indicate none)
		(3)  argument specification array, each element being an array of the form
				<arg name>  <arg type>
		(4)  result type, (#none or nil indicate none)
		(5)  array of category symbols, i.e. the categories in which this element should appear.
		(6)  help message. (optional)
		(7)  wording (optional)
		(8)  auto update flag (optional) - if #updating, set readout to refetch automatically

	Make new categories as needed.
	Consult Vocabulary class.initializeTestVocabulary for an example of use"
				
	| aMethodCategory aMethodInterface aSelector doc wording |
	aTable do:
		[:tuple |   tuple fifth do: [:aCategorySymbol |
			(aMethodCategory := self categoryAt: aCategorySymbol) ifNil: [ 
					aMethodCategory := ElementCategory new categoryName: aCategorySymbol.
					self addCategory: aMethodCategory].		
		
			aMethodInterface := MethodInterface new.
			aSelector := tuple first.
			aMethodInterface selector: aSelector type: tuple fourth setter: tuple second.
			aMethodCategory elementAt: aSelector put: aMethodInterface.
			self atKey: aSelector putMethodInterface: aMethodInterface.
			((tuple third ~~ #none) and: [tuple third isEmptyOrNil not])
				ifTrue:
					[aMethodInterface argumentVariables: (tuple third collect:
						[:pair | Variable new name: pair first type: pair second])].

		
			doc := (tuple size >= 6 and: [(#(nil none unused) includes: tuple sixth) not])
				ifTrue:
					[tuple sixth]
				ifFalse:
					[nil].
			wording := (tuple size >= 7 and: [(#(nil none unused) includes: tuple seventh) not])
				ifTrue:
					[tuple seventh]
				ifFalse:
					[aSelector].
			aMethodInterface
				wording: wording;
				helpMessage: doc.
			tuple size >= 8 ifTrue:
				[aMethodInterface setToRefetch]]].
! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 1/30/2001 16:31'!
allCategoryName
	"Answer the name by which the 'all' category is known.  This is redundant with two other things, including ClassOrganizer allCategory, at the moment -- should be cleaned up someday."

	^ '-- all --' asSymbol! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 2/24/2001 01:02'!
allMethodsInCategory: categoryName 
	"Answer a list of methods in the category of the given name"

	^ self allMethodsInCategory: categoryName forInstance: object ofClass: object class! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
allMethodsInCategory: categoryName forInstance: anObject ofClass: aClass
	"Answer a list of all methods in the etoy interface which are in the given category, on behalf of aClass and possibly anObject.  Note that there is no limitClass at play here."

	| aCategory |
	categoryName ifNil: [^ OrderedCollection new].
	categoryName = self allCategoryName ifTrue:
		[^ methodInterfaces collect: [:anInterface | anInterface selector]].

	aCategory := categories detect: [:cat | cat categoryName == categoryName asSymbol] ifNone: [^ OrderedCollection new].
	^ aCategory elementsInOrder collect: [:anElement | anElement selector] thenSelect:
			[:aSelector | aClass canUnderstand: aSelector]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 1/26/2001 16:22'!
allSelectorsInVocabulary
	"Answer a list of all selectors in the vocabulary"

	^ methodInterfaces collect: [:m | m selector]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 9/27/2001 17:47'!
asSymbol
	"Answer a symbol representing the receiver"

	^ self vocabularyName! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 9/13/2001 14:29'!
atKey: aKey putMethodInterface: anInterface
	"Place the given interface at the given key."

	methodInterfaces at: aKey put: anInterface! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/13/2000 20:02'!
categories
	"Answer a list of the categories in the receiver"

	^ categories! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/12/2000 06:13'!
categoriesContaining: aSelector forClass: aClass
	"Answer a list of categories that include aSelector"

	^ self categories select:
		[:aCategory | aCategory includesKey: aSelector]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/11/2000 15:31'!
categoryAt: aSymbol
	"Answer the category which has the given symbol as its categoryName, else nil if none found"

	^ categories detect: [:aCategory | aCategory categoryName == aSymbol] ifNone: [nil]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 1/30/2001 16:31'!
categoryCommentFor: aCategoryName
	"Answer diocumentation for the given category name, a symbol"

	categories do:
		[:cat | cat categoryName == aCategoryName ifTrue: [^ cat documentation]].

	aCategoryName = self allCategoryName ifTrue:
		[^ 'Shows all methods, whatever other categories they may belong to'].
	#(
	(all					'Danger!! An old designation that usually does NOT include all of anything!!')
	('as yet unclassified'	'Methods not yet given a specific classification in some class in which they are implemented')
	(private				'Methods that should only be called by self'))

		do:
			[:pair | pair first = aCategoryName ifTrue: [^ pair second]].

	^ aCategoryName, ' is a category that currently has no documentation'
! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryList
	"Answer the category list considering only code implemented in my 
	limitClass and lower. This variant is used when the limitClass and 
	targetObjct are known"
	| classToUse foundAMethod classThatImplements |
	classToUse := object class.
	^ categories
		select: [:aCategory | 
			foundAMethod := false.
			aCategory elementsInOrder
				do: [:aSpec | 
					classThatImplements := classToUse whichClassIncludesSelector: aSpec selector.
					(classThatImplements notNil
							and: [classThatImplements includesBehavior: limitClass])
						ifTrue: [foundAMethod := true]].
			foundAMethod]
		thenCollect: [:aCategory | aCategory categoryName]! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryListForInstance: targetObject ofClass: aClass limitClass: mostGenericClass 
	"Answer the category list for the given instance (may be nil) of the 
	given class, considering only code implemented in mostGenericClass and 
	lower "
	| classToUse foundAMethod classThatImplements |
	classToUse := targetObject
				ifNil: [aClass]
				ifNotNil: [targetObject class].
	^ categories
		select: [:aCategory | 
			foundAMethod := false.
			aCategory elementsInOrder
				do: [:aSpec | 
					classThatImplements := classToUse whichClassIncludesSelector: aSpec selector.
					(classThatImplements notNil
							and: [classThatImplements includesBehavior: mostGenericClass])
						ifTrue: [foundAMethod := true]].
			foundAMethod]
		thenCollect: [:aCategory | aCategory categoryName]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 9/12/2001 22:43'!
categoryWhoseTranslatedWordingIs: aWording
	"Answer the category whose translated is the one provided, or nil if none"

	^ self categories detect: [:aCategory | aCategory wording = aWording] ifNone: [nil]
! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryWithNameIn: categoryNames thatIncludesSelector: aSelector forInstance: targetInstance ofClass: targetClass
	"Answer the name of a category, from among the provided categoryNames, which defines the selector for the given class.  Note reimplementor"

	| itsName |
	self categories do:
		[:aCategory | ((categoryNames includes: (itsName := aCategory categoryName)) and:  [aCategory includesKey: aSelector])
			ifTrue:
				[^ itsName]].
	^ nil! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
categoryWordingAt: aSymbol
	"Answer the wording for the category at the given symbol"

	| result |
	result := self categoryAt: aSymbol.
	^result
		ifNotNil: [result wording]
		ifNil: [aSymbol]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/14/2000 05:55'!
classToUseFromInstance: anInstance ofClass: aClass
	"A small convenience to assist in complications arising because an instance is sometimes provided and sometimes not"

	^ anInstance ifNotNil: [anInstance class] ifNil: [aClass]
! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/12/2000 06:06'!
encompassesAPriori: aClass
	"Answer whether the receiver  a priori encompasses aClass -- see implementors"

	^ false! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 3/20/2001 15:42'!
includesDefinitionForSelector: aSelector
	"Answer whether the given selector is known to the vocabulary.  This is independent of whether its definition lies within the range specified by my limitClass.  Answer whether the given selector is known to the vocabulary.  Unsent at the moment, may disappear."

	^ methodInterfaces includesKey: aSelector! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/1/2000 21:57'!
includesSelector: aSelector
	"Answer whether the given selector is known to the vocabulary"

	^ methodInterfaces includesKey: aSelector! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
includesSelector: aSelector forInstance: anInstance ofClass: aTargetClass limitClass: mostGenericClass
	"Answer whether the vocabulary includes the given selector for the given class (and instance, if provided), only considering method implementations in mostGenericClass and lower"

	| classToUse aClass |

	(methodInterfaces includesKey: aSelector) ifFalse: [^ false].
	classToUse := self classToUseFromInstance: anInstance ofClass: aTargetClass.
	^ (aClass := classToUse whichClassIncludesSelector: aSelector)
		ifNil:
			[false]
		ifNotNil:
			[(aClass includesBehavior: mostGenericClass) and:
				[(self someCategoryThatIncludes: aSelector) notNil]]
! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initialize
	"Initialize the receiver (automatically called when instances are created via 'new')"

	super initialize.
	vocabularyName := #unnamed.
	categories := OrderedCollection new.
	methodInterfaces := IdentityDictionary new! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initializeFor: anObject
	"Initialize the receiver to bear a vocabulary suitable for anObject"

	object := anObject.
	vocabularyName := #unnamed.
	categories := OrderedCollection new.
	methodInterfaces := IdentityDictionary new.
	self documentation: 'A vocabulary that has not yet been documented'.
! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
initializeFromTable: aTable
	"Initialize the receiver from a list of method-specification tuples, each of the form:
		(1)	selector
		(2)	companion setter selector (#none or nil indicate none)
		(3)  argument specification array, each element being an array of the form
				<arg name>  <arg type>
		(4)  result type, (#none or nil indicate none)
		(5)  array of category symbols, i.e. the categories in which this element should appear.
		(6)  help message. (optional)
		(7)  wording (optional)
		(8)  auto update flag (optional) - if #updating, set readout to refetch automatically

	Consult Vocabulary class.initializeTestVocabulary for an example of use"
				
	|  aMethodCategory categoryList aMethodInterface aSelector doc wording |
	categoryList := Set new.
	aTable do:
		[:tuple | categoryList addAll: tuple fifth].
	categoryList := categoryList asSortedArray.
	categoryList do:
		[:aCategorySymbol |
			aMethodCategory := ElementCategory new categoryName: aCategorySymbol.
			aTable do:
				[:tuple | (tuple fifth includes: aCategorySymbol) ifTrue:
					[aMethodInterface := MethodInterface new.
					aSelector := tuple first.
					aMethodInterface selector: aSelector type: tuple fourth setter: tuple second.
					aMethodCategory elementAt: aSelector put: aMethodInterface.
					self atKey: aSelector putMethodInterface: aMethodInterface.
					((tuple third ~~ #none) and: [tuple third isEmptyOrNil not])
						ifTrue:
							[aMethodInterface argumentVariables: (tuple third collect:
								[:pair | Variable new name: pair first type: pair second])].
					doc := (tuple size >= 6 and: [(#(nil none unused) includes: tuple sixth) not])
						ifTrue:
							[tuple sixth]
						ifFalse:
							[nil].
 					wording := (tuple size >= 7 and: [(#(nil none unused) includes: tuple seventh) not])
						ifTrue:
							[tuple seventh]
						ifFalse:
							[aSelector asString].
					aMethodInterface
						wording: wording;
						helpMessage: doc.
					tuple size >= 8 ifTrue:
						[aMethodInterface setToRefetch]]].
			self addCategory: aMethodCategory]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 2/23/2001 23:30'!
methodInterfaceAt: aSelector ifAbsent: aBlock
	"Answer the vocabulary's method interface for the given selector; if absent, return the result of evaluating aBlock"

	^ methodInterfaces at: aSelector ifAbsent: [aBlock value]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 8/11/2002 02:25'!
methodInterfaceForSelector: aSelector class: aClass
	"Answer a method interface for the selector"

	^ self methodInterfaceAt: aSelector ifAbsent:
		[MethodInterface new conjuredUpFor: aSelector class: aClass]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 5/25/2001 10:42'!
methodInterfacesDo: aBlock
	"Evaluate aBlock on behalf, in turn, of each of my methodInterfaces"

	methodInterfaces do: aBlock

	! !

!Vocabulary methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:39'!
methodInterfacesInCategory: categoryName forInstance: anObject ofClass: aClass limitClass: aLimitClass
	"Answer a list of method interfaces of all methods in the given category, provided they are implemented no further away than aLimitClass."

	| defClass |
	^ ((self allMethodsInCategory: categoryName forInstance: anObject ofClass: aClass) collect:
		[:sel | methodInterfaces at: sel ifAbsent:
			[MethodInterface new conjuredUpFor: sel class: aClass]]) select:
				[:aMethodInterface |
					defClass := aClass whichClassIncludesSelector: aMethodInterface selector.
					(defClass notNil and: [defClass includesBehavior: aLimitClass])]! !

!Vocabulary methodsFor: 'method list' stamp: 'sw 11/13/2001 09:34'!
phraseSymbolsToSuppress
	"Answer a dictatorially-imposed list of phrase-symbols that are to be suppressed from viewers, even if they otherwise show up.  Note that EToyVocabulary reimplements"

	^ #()! !

!Vocabulary methodsFor: 'printing' stamp: 'sw 12/1/2000 22:05'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that   identifies the receiver."

	super printOn: aStream.
	vocabularyName ifNotNil: [aStream nextPutAll: ' named "', vocabularyName, '"']! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
renameCategoryFrom: oldName to: newName
	"Rename the category currently known by oldName to be newName.  No senders at present but once a UI is establshed for renaming categories, this will be useful."

	| aCategory |
	(aCategory := self categoryAt: oldName) ifNil: [^ self].
	aCategory categoryName: newName! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 9/27/2001 03:24'!
representsAType
	"Answer whether this vocabulary represents an end-user-sensible data type"

	^ false! !

!Vocabulary methodsFor: 'private' stamp: 'mir 7/15/2004 19:27'!
setCategoryStrings: categoryTriplets
	"Establish the category strings as per (internalCategorySymbol newCategoryWording balloon-help)"

	| category |
	categoryTriplets do:
		[:triplet |
			(category := self categoryAt: triplet first) ifNotNil: [
				category wording: triplet second.
				category helpMessage: triplet third]]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 3/19/2001 23:55'!
someCategoryThatIncludes: aSelector
	"Answer the name of a category that includes the selector, nil if none"

	^ categories detect: [:c | c includesKey: aSelector] ifNone: [nil]! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
strings
	| strm |
	"Get started making a vocabulary for a foreign language.  That is, build a method like #addGermanVocabulary, but for another language.  
	Returns this vocabulary in the same form used as the input used for foreign languages.  To avoid string quote problems, execute
	Transcript show: Vocabulary eToyVocabulary strings.
and copy the text from the transcript to the method you are building."

	"selector		wording			documentation"

strm := WriteStream on: (String new: 400).
methodInterfaces keys asSortedCollection do: [:sel |
	strm cr; nextPut: $(;
		nextPutAll: sel; tab; tab; tab; nextPut: $';
		nextPutAll: (methodInterfaces at: sel) wording;
		nextPut: $'; tab; tab; tab; nextPut: $';
		nextPutAll: (methodInterfaces at: sel) documentation;
		nextPut: $'; nextPut: $)].
^ strm contents! !

!Vocabulary methodsFor: 'translation' stamp: 'tak 8/8/2005 12:52'!
translatedWordingFor: aSymbol
	"If I have a translated wording for aSymbol, return it, else return aSymbol.  Caveat: at present, this mechanism is only germane for *assignment-operator wordings*"

	#(: Incr: Decr: Mult:) with: #('' 'increase by' 'decrease by' 'multiply by') do:
		[:a :b | aSymbol = a ifTrue: [^ b translated]].

	^ aSymbol translated! !

!Vocabulary methodsFor: 'translation' stamp: 'sw 5/25/2001 10:45'!
translatedWordingsFor: symbolList
	"Answer a list giving the translated wordings for the input list. Caveat: at present, this mechanism is only germane for *categories*"

	^ symbolList collect: [:sym | self translatedWordingFor: sym]
! !

!Vocabulary methodsFor: 'translation' stamp: 'yo 7/2/2004 21:51'!
translationKeyFor: translatedWording

	self flag: #yo.
	^ translatedWording.
! !

!Vocabulary methodsFor: 'translation' stamp: 'mir 7/13/2004 00:42'!
translationTable
	^translationTable ifNil: [ElementCategory new]! !

!Vocabulary methodsFor: 'queries' stamp: 'sw 12/1/2000 22:05'!
vocabularyName
	"Answer the name of the vocabulary"

	^ vocabularyName! !

!Vocabulary methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:39'!
vocabularyName: aName
	"Set the name of the vocabulary as indicated"

	vocabularyName := aName! !
